'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _ = require('lodash');
var axios = _interopDefault(require('axios'));
var Vue = _interopDefault(require('vue'));

var asyncGenerator = function () {
  function AwaitValue(value) {
    this.value = value;
  }

  function AsyncGenerator(gen) {
    var front, back;

    function send(key, arg) {
      return new Promise(function (resolve, reject) {
        var request = {
          key: key,
          arg: arg,
          resolve: resolve,
          reject: reject,
          next: null
        };

        if (back) {
          back = back.next = request;
        } else {
          front = back = request;
          resume(key, arg);
        }
      });
    }

    function resume(key, arg) {
      try {
        var result = gen[key](arg);
        var value = result.value;

        if (value instanceof AwaitValue) {
          Promise.resolve(value.value).then(function (arg) {
            resume("next", arg);
          }, function (arg) {
            resume("throw", arg);
          });
        } else {
          settle(result.done ? "return" : "normal", result.value);
        }
      } catch (err) {
        settle("throw", err);
      }
    }

    function settle(type, value) {
      switch (type) {
        case "return":
          front.resolve({
            value: value,
            done: true
          });
          break;

        case "throw":
          front.reject(value);
          break;

        default:
          front.resolve({
            value: value,
            done: false
          });
          break;
      }

      front = front.next;

      if (front) {
        resume(front.key, front.arg);
      } else {
        back = null;
      }
    }

    this._invoke = send;

    if (typeof gen.return !== "function") {
      this.return = undefined;
    }
  }

  if (typeof Symbol === "function" && Symbol.asyncIterator) {
    AsyncGenerator.prototype[Symbol.asyncIterator] = function () {
      return this;
    };
  }

  AsyncGenerator.prototype.next = function (arg) {
    return this._invoke("next", arg);
  };

  AsyncGenerator.prototype.throw = function (arg) {
    return this._invoke("throw", arg);
  };

  AsyncGenerator.prototype.return = function (arg) {
    return this._invoke("return", arg);
  };

  return {
    wrap: function (fn) {
      return function () {
        return new AsyncGenerator(fn.apply(this, arguments));
      };
    },
    await: function (value) {
      return new AwaitValue(value);
    }
  };
}();





var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();





var defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};

var get$1 = function get$$1(object, property, receiver) {
  if (object === null) object = Function.prototype;
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent === null) {
      return undefined;
    } else {
      return get$$1(parent, property, receiver);
    }
  } else if ("value" in desc) {
    return desc.value;
  } else {
    var getter = desc.get;

    if (getter === undefined) {
      return undefined;
    }

    return getter.call(receiver);
  }
};

var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};











var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

var Response = function () {
    function Response(response) {
        classCallCheck(this, Response);

        this.response = response;
    }

    createClass(Response, [{
        key: 'getData',
        value: function getData() {
            return _.get(this.response, 'data', null);
        }
    }, {
        key: 'getStatus',
        value: function getStatus() {
            return _.get(this.response, 'status');
        }
    }, {
        key: 'getHeaders',
        value: function getHeaders() {
            return _.get(this.response, 'headers', {});
        }
    }, {
        key: 'getValidationErrors',
        value: function getValidationErrors() {
            return _.get(this.response, 'data', null);
        }
    }]);
    return Response;
}();

var RequestError = function () {
    function RequestError(error, response) {
        classCallCheck(this, RequestError);

        this.error = error;
        this.response = response;
        this.stack = new Error().stack;
        this.message = error.message;
    }

    createClass(RequestError, [{
        key: "toString",
        value: function toString() {
            return this.message;
        }
    }, {
        key: "getError",
        value: function getError() {
            return this.error;
        }
    }, {
        key: "getResponse",
        value: function getResponse() {
            return this.response;
        }
    }]);
    return RequestError;
}();

var Request = function () {
    function Request(config) {
        classCallCheck(this, Request);

        this.config = config;
    }

    /**
     * @returns {Promise}
     */


    createClass(Request, [{
        key: 'send',
        value: function send() {
            var _this = this;

            return new Promise(function (resolve, reject) {
                axios.request(_this.config).then(function (response) {
                    return resolve(new Response(response));
                }).catch(function (error) {
                    return reject(new RequestError(error, new Response(error.response)));
                });
            });
        }
    }]);
    return Request;
}();

/**
 * Binds all methods of a class instance to itself.
 */
var autobind = function autobind(instance) {
    for (var obj = instance; obj; obj = Object.getPrototypeOf(obj)) {

        // We're the end of the inheritance chain if we've reached 'Object'.
        if (obj.constructor.name === 'Object') {
            return;
        }

        var names = Object.getOwnPropertyNames(obj);

        // Bind each function to the instance.
        for (var i = 0; i < names.length; i++) {
            var name = names[i];

            // No need to bind getters, as attempting to access them would also
            // invoke them which is something we don't want to do here.
            if (Object.getOwnPropertyDescriptor(obj, name).get) {
                continue;
            }

            // We're using `defineProperty` here so that we don't make all the
            // class methods enumerable when we replace them.
            if (typeof obj[name] === 'function' && name !== 'constructor') {
                Object.defineProperty(instance, name, {
                    value: instance[name].bind(instance),
                    enumerable: false,
                    configurable: true,
                    writable: true
                });
            }
        }
    }
};

/**
 * Base class for all things common between Model and Collection.
 */

var Base = function () {
    function Base(options) {
        classCallCheck(this, Base);

        autobind(this);

        // Define an automatic unique ID. This is primarily to distinguish
        // between multiple instances of the same name and data.
        Object.defineProperty(this, '_uid', {
            value: _.uniqueId(),
            enumerable: false,
            configurable: false,
            writable: false
        });

        Vue.set(this, '_listeners', {}); // Event listeners
        Vue.set(this, '_options', {}); // Internal option store

        this.setOptions(options);
        this.boot();
    }

    /**
     * @returns {string} The class name of this instance.
     */


    createClass(Base, [{
        key: 'boot',


        /**
         * Called after construction, this hook allows you to add some extra setup
         * logic without having to override the constructor.
         */
        value: function boot() {}

        /**
         * Returns a route configuration in the form {key: name}, where key may be
         * 'save', 'fetch', 'delete' or any other custom key, and the name is what
         * will be passed to the route resolver to generate the URL. See @getURL
         *
         * @returns {Object}
         */

    }, {
        key: 'routes',
        value: function routes() {
            return {};
        }

        /**
         * Returns the default context for all events emitted by this instance.
         *
         * @returns {Object}
         */

    }, {
        key: 'getDefaultEventContext',
        value: function getDefaultEventContext() {
            return { target: this };
        }

        /**
         * @returns {string} Default string representation.
         */

    }, {
        key: 'toString',
        value: function toString() {
            return '<' + this.$class + ' #' + this._uid + '>';
        }

        /**
         * Emits an event by name to all registered listeners on that event.
          * Listeners will be called in the order that they were added. If a listener
         * returns `false`, no other listeners will be called.
         *
         * @param {string} event    The name of the event to emit.
         * @param {Object} context  The context of the event, passed to listeners.
         */

    }, {
        key: 'emit',
        value: function emit(event) {
            var context = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

            var listeners = _.get(this._listeners, event);

            if (!listeners) {
                return;
            }

            // Create the context for the event.
            context = _.defaults({}, context, this.getDefaultEventContext());

            // Run through each listener. If any of them return false, stop the
            // iteration and mark that the event wasn't handled by all listeners.
            _.each(listeners, function (listener) {
                return listener(context);
            });
        }

        /**
         * Registers an event listener for a given event.
         *
         * Event names can be comma-separated to register multiple events.
         *
         * @param {string}   event      The name of the event to listen for.
         * @param {function} listener   The event listener, accepts context.
         */

    }, {
        key: 'on',
        value: function on(event, listener) {
            var _this = this;

            var events = _.map(_.split(event, ','), _.trim);

            _.each(events, function (event) {
                _this._listeners[event] = _this._listeners[event] || [];
                _this._listeners[event].push(listener);
            });
        }

        /**
         * @returns {Object} Parameters to use for replacement in route patterns.
         */

    }, {
        key: 'getRouteParameters',
        value: function getRouteParameters() {
            return {};
        }

        /**
         * @returns {RegExp|string} Pattern to match and group route parameters.
         */

    }, {
        key: 'getRouteParameterPattern',
        value: function getRouteParameterPattern() {
            return this.getOption('routeParameterPattern');
        }

        /**
         * @returns {RegExp} The default route parameter pattern.
         */

    }, {
        key: 'getDefaultRouteParameterPattern',
        value: function getDefaultRouteParameterPattern() {
            return (/\{([^}]+)\}/
            );
        }

        /**
         * @returns {Object} This class' default options.
         */

    }, {
        key: 'getDefaultOptions',
        value: function getDefaultOptions() {
            return {

                // Default HTTP methods for requests.
                methods: this.getDefaultMethods(),

                // Default route parameter interpolation pattern.
                routeParameterPattern: this.getDefaultRouteParameterPattern(),

                // The HTTP status code to use for indicating a validation error.
                validationErrorStatus: 422
            };
        }

        /**
         * @param {Array|string} path     Option path resolved by `_.get`
         * @param {*}            fallback Fallback value if the option is not set.
         *
         * @returns {*} The value of the given option path.
         */

    }, {
        key: 'getOption',
        value: function getOption(path) {
            var fallback = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

            return _.get(this._options, path, fallback);
        }

        /**
         * @returns {Object} This instance's default options.
         */

    }, {
        key: 'options',
        value: function options() {
            return {};
        }

        /**
         * Sets an option.
         *
         * @param {string} path
         * @param {*}      value
         */

    }, {
        key: 'setOption',
        value: function setOption(path, value) {
            _.set(this._options, path, value);
        }

        /**
         * Sets all given options. Successive values for the same option won't be
         * overwritten, so this follows the 'defaults' behaviour, and not 'merge'.
         *
         * @param {...Object} options One or more objects of options.
         */

    }, {
        key: 'setOptions',
        value: function setOptions() {
            for (var _len = arguments.length, options = Array(_len), _key = 0; _key < _len; _key++) {
                options[_key] = arguments[_key];
            }

            Vue.set(this, '_options', _.defaultsDeep.apply(_, [{}].concat(options, [// Given options
            this.options(), // Instance defaults
            this.getDefaultOptions() // Class defaults
            ])));
        }

        /**
         * Returns all the options that are currently set on this instance.
         *
         * @return {Object}
         */

    }, {
        key: 'getOptions',
        value: function getOptions() {
            return _.defaultTo(this._options, {});
        }

        /**
         * Returns a function that translates a route key and parameters to a URL.
         *
         * @returns {Function} Will be passed `route` and `parameters`
         */

    }, {
        key: 'getRouteResolver',
        value: function getRouteResolver() {
            return this.getDefaultRouteResolver();
        }

        /**
         * @returns {Object} An object consisting of all route string replacements.
         */

    }, {
        key: 'getRouteReplacements',
        value: function getRouteReplacements(route) {
            var parameters = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

            var replace$$1 = {};
            var pattern = new RegExp(this.getRouteParameterPattern(), 'g');

            for (var parameter; (parameter = pattern.exec(route)) !== null;) {
                replace$$1[parameter[0]] = parameters[parameter[1]];
            }

            return replace$$1;
        }

        /**
         * Returns the default URL provider, which assumes that route keys are URL's,
         * and parameter replacement syntax is in the form "{param}".
         *
         * @returns {Function}
         */

    }, {
        key: 'getDefaultRouteResolver',
        value: function getDefaultRouteResolver() {
            var _this2 = this;

            return function (route) {
                var parameters = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

                var replacements = _this2.getRouteReplacements(route, parameters);

                // Replace all route parameters with their replacement values.
                return _.reduce(replacements, function (result, value, parameter) {
                    return _.replace(result, parameter, value);
                }, route);
            };
        }

        /**
         * @returns {Object} The data to send to the server when saving this model.
         */

    }, {
        key: 'getDeleteBody',
        value: function getDeleteBody() {
            return {};
        }

        /**
         * @returns {Object} Query parameters that will be appended to the `fetch` URL.
         */

    }, {
        key: 'getFetchQuery',
        value: function getFetchQuery() {
            return {};
        }

        /**
         * @returns {Object} Query parameters that will be appended to the `save` URL.
         */

    }, {
        key: 'getSaveQuery',
        value: function getSaveQuery() {
            return {};
        }

        /**
         * @returns {Object} Query parameters that will be appended to the `delete` URL.
         */

    }, {
        key: 'getDeleteQuery',
        value: function getDeleteQuery() {
            return {};
        }

        /**
         * @returns {string} The key to use when generating the `fetch` URL.
         */

    }, {
        key: 'getFetchRoute',
        value: function getFetchRoute() {
            return this.getRoute('fetch');
        }

        /**
         * @returns {string} The key to use when generating the `save` URL.
         */

    }, {
        key: 'getSaveRoute',
        value: function getSaveRoute() {
            return this.getRoute('save');
        }

        /**
         * @returns {string} The key to use when generating the `delete` URL.
         */

    }, {
        key: 'getDeleteRoute',
        value: function getDeleteRoute() {
            return this.getRoute('delete');
        }

        /**
         * @returns {Object} Headers to use when making a save request.
         */

    }, {
        key: 'getSaveHeaders',
        value: function getSaveHeaders() {
            return {};
        }

        /**
         * @returns {Object} Headers to use when making any request.
         */

    }, {
        key: 'getDefaultHeaders',
        value: function getDefaultHeaders() {
            return {};
        }

        /**
         * @returns {Object} Headers to use when making a fetch request.
         */

    }, {
        key: 'getFetchHeaders',
        value: function getFetchHeaders() {
            return {};
        }

        /**
         * @returns {Object} Headers to use when making a delete request.
         */

    }, {
        key: 'getDeleteHeaders',
        value: function getDeleteHeaders() {
            return {};
        }

        /**
         * @returns {Object} Default HTTP methods.
         */

    }, {
        key: 'getDefaultMethods',
        value: function getDefaultMethods() {
            return {
                fetch: 'GET',
                save: 'POST',
                update: 'POST',
                create: 'POST',
                patch: 'PATCH',
                delete: 'DELETE'
            };
        }

        /**
         * @returns {string} HTTP method to use when making a save request.
         */

    }, {
        key: 'getSaveMethod',
        value: function getSaveMethod() {
            return this.getOption('methods.save');
        }

        /**
         * @returns {string} HTTP method to use when making a fetch request.
         */

    }, {
        key: 'getFetchMethod',
        value: function getFetchMethod() {
            return this.getOption('methods.fetch');
        }

        /**
         * @returns {string} HTTP method to use when updating a resource.
         */

    }, {
        key: 'getUpdateMethod',
        value: function getUpdateMethod() {
            return this.getOption('methods.update');
        }

        /**
         * @returns {string} HTTP method to use when patching a resource.
         */

    }, {
        key: 'getPatchMethod',
        value: function getPatchMethod() {
            return this.getOption('methods.patch');
        }

        /**
         * @returns {string} HTTP method to use when creating a resource.
         */

    }, {
        key: 'getCreateMethod',
        value: function getCreateMethod() {
            return this.getOption('methods.create');
        }

        /**
         * @returns {string} HTTP method to use when deleting a resource.
         */

    }, {
        key: 'getDeleteMethod',
        value: function getDeleteMethod() {
            return this.getOption('methods.delete');
        }

        /**
         * @returns {number} The HTTP status code that indicates a validation error.
         */

    }, {
        key: 'getValidationErrorStatus',
        value: function getValidationErrorStatus() {
            return _.defaultTo(this.getOption('validationErrorStatus'), 422);
        }

        /**
         * @returns {boolean} `true` if the response indicates a validation error.
         */

    }, {
        key: 'isBackendValidationError',
        value: function isBackendValidationError(error) {

            // The error must have a response for it to be a validation error.
            if (!_.invoke(error, 'getResponse', false)) {
                return false;
            }

            var status = error.getResponse().getStatus();
            var invalid = this.getValidationErrorStatus();

            return status == invalid;
        }

        /**
         * @return {string|undefined} Route value by key.
         */

    }, {
        key: 'getRoute',
        value: function getRoute(key, fallback) {
            var route = _.get(this.routes(), key, _.get(this.routes(), fallback));

            if (!route) {
                throw new Error('Invalid or missing route');
            }

            return route;
        }

        /**
         * @returns {string} The full URL to use when making a fetch request.
         */

    }, {
        key: 'getFetchURL',
        value: function getFetchURL() {
            return this.getURL(this.getFetchRoute(), this.getRouteParameters());
        }

        /**
         * @returns {string} The full URL to use when making a save request.
         */

    }, {
        key: 'getSaveURL',
        value: function getSaveURL() {
            return this.getURL(this.getSaveRoute(), this.getRouteParameters());
        }

        /**
         * @returns {string} The full URL to use when making a delete request.
         */

    }, {
        key: 'getDeleteURL',
        value: function getDeleteURL() {
            return this.getURL(this.getDeleteRoute(), this.getRouteParameters());
        }

        /**
         * @param {string} route      The route key to use to generate the URL.
         * @param {Object} parameters Route parameters.
         *
         * @returns {string} A URL that was generated using the given route key.
         */

    }, {
        key: 'getURL',
        value: function getURL(route) {
            var parameters = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

            return this.getRouteResolver()(route, parameters);
        }

        /**
         * @returns {Request} A new `Request` using the given configuration.
         */

    }, {
        key: 'getRequest',
        value: function getRequest(config) {
            return new Request(config);
        }

        /**
         * This is the central component for all HTTP requests and handling.
         *
         * @param  {Object}     config      Request configuration
         * @param  {function}   onRequest   Called before the request is made.
         * @param  {function}   onSuccess   Called when the request was successful.
         * @param  {function}   onFailure   Called when the request failed.
         */

    }, {
        key: 'request',
        value: function request(config, onRequest, onSuccess, onFailure) {
            var _this3 = this;

            return new Promise(function (resolve, reject) {

                var check = onRequest();

                // Request should be skipped but the promise should not be resolved.
                if (check === false) {
                    return;
                }

                // Request should be skipped but should be considered successful.
                if (check === true) {
                    onSuccess(null);
                    return resolve(null);
                }

                // Support passing the request configuration as a function, to allow
                // for deferred resolution of certain values that may have changed
                // during the call to "onRequest".
                if (_.isFunction(config)) {
                    config = config();
                }

                // Apply the default headers.
                _.defaults(config.headers, _this3.getDefaultHeaders());

                // Make the request.
                return _this3.getRequest(config).send()

                // Success
                .then(function (response) {
                    onSuccess(response);
                    return resolve(response);
                })

                // Failure
                .catch(function (error) {
                    onFailure(error);
                    return reject(error);
                })

                // Failure fallback, for errors that occur in `onFailure`.
                .catch(function (error) {
                    return reject(error);
                });
            });
        }

        /**
         * Fetches data from the database/API.
         *
         * @returns {Promise}
         */

    }, {
        key: 'fetch',
        value: function fetch() {
            var _this4 = this;

            var config = function config() {
                return {
                    url: _this4.getFetchURL(),
                    method: _this4.getFetchMethod(),
                    params: _this4.getFetchQuery(),
                    headers: _this4.getFetchHeaders()
                };
            };

            return this.request(config, this.onFetch, this.onFetchSuccess, this.onFetchFailure);
        }

        /**
         * Persists data to the database/API.
         * @returns {Promise}
         */

    }, {
        key: 'save',
        value: function save() {
            var _this5 = this;

            var config = function config() {
                return {
                    url: _this5.getSaveURL(),
                    method: _this5.getSaveMethod(),
                    data: _this5.getSaveData(),
                    params: _this5.getSaveQuery(),
                    headers: _this5.getSaveHeaders()
                };
            };

            return this.request(config, this.onSave, this.onSaveSuccess, this.onSaveFailure);
        }

        /**
         * Removes model or collection data from the database/API.
         * @returns {Promise}
         */

    }, {
        key: 'delete',
        value: function _delete() {
            var _this6 = this;

            var config = function config() {
                return {
                    url: _this6.getDeleteURL(),
                    method: _this6.getDeleteMethod(),
                    data: _this6.getDeleteBody(),
                    params: _this6.getDeleteQuery(),
                    headers: _this6.getDeleteHeaders()
                };
            };

            return this.request(config, this.onDelete, this.onDeleteSuccess, this.onDeleteFailure);
        }
    }, {
        key: '$class',
        get: function get$$1() {
            return Object.getPrototypeOf(this).constructor.name;
        }
    }]);
    return Base;
}();

var ResponseError = function () {
    function ResponseError(message, response) {
        classCallCheck(this, ResponseError);

        this.message = message;
        this.response = response;
        this.stack = new Error().stack;
    }

    createClass(ResponseError, [{
        key: "toString",
        value: function toString() {
            return this.message;
        }
    }, {
        key: "getResponse",
        value: function getResponse() {
            return this.response;
        }
    }]);
    return ResponseError;
}();

var ValidationError = function () {
    function ValidationError(errors) {
        var message = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'Model did not pass validation';
        classCallCheck(this, ValidationError);

        this.message = message;
        this.errors = errors;
        this.stack = new Error().stack;
    }

    createClass(ValidationError, [{
        key: 'toString',
        value: function toString() {
            return this.message;
        }
    }, {
        key: 'getValidationErrors',
        value: function getValidationErrors() {
            return this.errors;
        }
    }]);
    return ValidationError;
}();

var ProxyResponse = function () {
    function ProxyResponse(status) {
        var data = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
        var headers = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
        classCallCheck(this, ProxyResponse);

        this.data = _.defaultTo(data, {});
        this.headers = _.defaultTo(headers, {});
        this.status = _.toSafeInteger(status);
    }

    createClass(ProxyResponse, [{
        key: 'getData',
        value: function getData() {
            return this.data;
        }
    }, {
        key: 'getStatus',
        value: function getStatus() {
            return this.status;
        }
    }, {
        key: 'getHeaders',
        value: function getHeaders() {
            return this.headers;
        }
    }, {
        key: 'getValidationErrors',
        value: function getValidationErrors() {
            return this.data;
        }
    }]);
    return ProxyResponse;
}();

/**
 * Used as a marker to indicate that pagination is not enabled.
 */
var NO_PAGE = null;

/**
 * Used as a marker to indicate that a collection has paged through all results.
 */
var LAST_PAGE = 0;

/**
 * Base collection class.
 */

var Collection = function (_Base) {
    inherits(Collection, _Base);
    createClass(Collection, [{
        key: 'length',


        /**
         * Accessor to support Array.length semantics.
         */
        get: function get$$1() {
            return this.size();
        }

        /**
         * Creates a new instance, called when using 'new'.
         *
         * @param  {Array}  [models]    Models to add to this collection.
         * @param  {Object} [options]   Extra options to set on this collection.
         */

    }]);

    function Collection() {
        var models = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
        var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
        var attributes = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
        classCallCheck(this, Collection);

        var _this = possibleConstructorReturn(this, (Collection.__proto__ || Object.getPrototypeOf(Collection)).call(this, options));

        Vue.set(_this, 'models', []); // Model store.
        Vue.set(_this, '_attributes', {}); // Property store.
        Vue.set(_this, '_registry', {}); // Model registry.
        Vue.set(_this, '_page', NO_PAGE);

        _this.clearState();

        // Set all given attributes.
        _this.set(_.defaultsDeep({}, attributes, _this.defaults()));

        // Add all given models (if any) to this collection. We explicitly ask
        // for the values here as it's common for some sources to be objects.
        if (models) {
            _this.add(_.values(models));
        }
        return _this;
    }

    /**
     * Creates a copy of this collection. Model references are preserved so
     * changes to the models inside the clone will also affect the subject.
     *
     * @returns {Collection}
     */


    createClass(Collection, [{
        key: 'clone',
        value: function clone() {
            var clone = new this.constructor();

            // Add all the existing models.
            clone.add(this.models);

            // Make sure that the clone has the same existing options.
            clone.setOptions(this.getOptions());

            return clone;
        }

        /**
         * @return {Model} The class/constructor for this collection's model type.
         */

    }, {
        key: 'model',
        value: function model() {
            return this.getOption('model');
        }

        /**
         * @return {Object} Default attributes
         */

    }, {
        key: 'defaults',
        value: function defaults$$1() {
            return {};
        }

        /**
         * @return {*} The value of an attribute, or a given fallback if not set.
         */

    }, {
        key: 'get',
        value: function get$$1(attribute, fallback) {
            return _.get(this._attributes, attribute, fallback);
        }

        /**
         * Sets an attribute's value, or an object of attributes.
         *
         * @param {string|Object} attribute
         * @param {*}             value
         */

    }, {
        key: 'set',
        value: function set$$1(attribute, value) {
            var _this2 = this;

            if (_.isPlainObject(attribute)) {
                _.each(attribute, function (value, key) {
                    _this2.set(key, value);
                });

                return;
            }

            Vue.set(this._attributes, attribute, value);
        }

        /**
         * Returns the default options for this model.
         *
         * @returns {Object}
         */

    }, {
        key: 'getDefaultOptions',
        value: function getDefaultOptions() {
            return _.merge(get$1(Collection.prototype.__proto__ || Object.getPrototypeOf(Collection.prototype), 'getDefaultOptions', this).call(this), {

                // The class/constructor for this collection's model type.
                model: Model,

                // Whether this collection should send model identifiers as JSON
                // in the body of a delete request, instead of a query parameter.
                useDeleteBody: true
            });
        }

        /**
         * @returns {Object} Parameters to use for replacement in route patterns.
         */

    }, {
        key: 'getRouteParameters',
        value: function getRouteParameters() {
            return _.merge({}, get$1(Collection.prototype.__proto__ || Object.getPrototypeOf(Collection.prototype), 'getRouteParameters', this).call(this), this._attributes, {
                page: this._page
            });
        }

        /**
         * Removes all errors from the models in this collection.
         */

    }, {
        key: 'clearErrors',
        value: function clearErrors() {
            _.each(this.models, _.method('clearErrors'));
        }

        /**
         * Resets model state, ie. `loading`, etc back to their initial states.
         */

    }, {
        key: 'clearState',
        value: function clearState() {
            Vue.set(this, 'loading', true);
            Vue.set(this, 'saving', false);
            Vue.set(this, 'deleting', false);
            Vue.set(this, 'fatal', false);
        }

        /**
         * Removes all models from this collection.
         */

    }, {
        key: 'clearModels',
        value: function clearModels() {
            var _this3 = this;

            var models = this.models;

            // Clear the model store, but keep a reference.
            Vue.set(this, 'models', []);

            // Notify each model that it has been removed from this collection.
            _.each(models, function (model) {
                _this3.onRemove(model);
            });
        }

        /**
         * Removes all models from this collection.
         */

    }, {
        key: 'clear',
        value: function clear() {
            this.clearModels();
            this.clearState();
        }

        /**
         * Syncs all models in this collection. This method delegates to each model
         * so follows the same signature and effects as `Model.sync`.
         */

    }, {
        key: 'sync',
        value: function sync() {
            _.each(this.models, _.method('sync'));
        }

        /**
         * Resets all models in this collection. This method delegates to each model
         * so follows the same signature and effects as `Model.reset`.
         *
         * @param {string|string[]} attribute
         */

    }, {
        key: 'reset',
        value: function reset() {
            for (var _len = arguments.length, attribute = Array(_len), _key = 0; _key < _len; _key++) {
                attribute[_key] = arguments[_key];
            }

            _.each(this.models, _.method.apply(_, ['reset'].concat(attribute)));
        }

        /**
         * Returns the number of models in this collection.
         */

    }, {
        key: 'size',
        value: function size$$1() {
            return _.size(this.models);
        }

        /**
         * @returns {boolean} `true` if the collection is empty, `false` otherwise.
         */

    }, {
        key: 'isEmpty',
        value: function isEmpty$$1() {
            return this.size() === 0;
        }

        /**
         * @returns {Object} A native representation of this collection that will
         *                   determine the contents of JSON.stringify(collection).
         */

    }, {
        key: 'toJSON',
        value: function toJSON() {
            return this.models;
        }

        /**
         * @returns {bool} Whether all models in this collection have valid data.
         */

    }, {
        key: 'validate',
        value: function validate() {
            return _.reduce(this.models, function (valid, model) {
                return model.validate() && valid;
            }, true);
        }

        /**
         * Create a new model of this collection's model type.
         *
         * @param {Object} attributes
         *
         * @returns {Model} A new instance of this collection's model.
         */

    }, {
        key: 'createModel',
        value: function createModel(attributes) {
            return new (this.model())(attributes);
        }

        /**
         * Removes a model from the model registry.
         *
         * @param {Model} model
         */

    }, {
        key: 'removeModelFromRegistry',
        value: function removeModelFromRegistry(model) {
            _.unset(this._registry, model._uid);
        }

        /**
         * @return {Boolean} true if this collection has the model in its registry.
         */

    }, {
        key: 'hasModelInRegistry',
        value: function hasModelInRegistry(model) {
            return _.has(this._registry, model._uid);
        }

        /**
         * Adds a model from the model registry.
         *
         * @param {Model} model
         */

    }, {
        key: 'addModelToRegistry',
        value: function addModelToRegistry(model) {
            _.set(this._registry, model._uid, 1);
        }

        /**
         * Called when a model has been added to this collection.
         *
         * @param {Model} model
         */

    }, {
        key: 'onAdd',
        value: function onAdd(model) {
            model.registerCollection(this);
            this.addModelToRegistry(model);
            this.emit('add', { model: model });
        }

        /**
         * Adds a model to this collection.
         *
         * This method returns a single model if only one was given, but will return
         * an array of all added models if an array was given.
         *
         * @param {Model|Array|Object} model Adds a model instance or plain object,
         *                                   or an array of either, to this collection.
         *                                   A model instance will be created and
         *                                   returned if passed a plain object.
         *
         * @returns {Model|Array} The added model or array of added models.
         */

    }, {
        key: 'add',
        value: function add() {
            var model = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};


            // If given an array, assume an array of models and add them all.
            if (_.isArray(model)) {
                return _.filter(_.map(model, this.add));
            }

            // Objects should be converted to model instances first, then added.
            if (_.isPlainObject(model)) {
                return this.add(this.createModel(model));
            }

            // This is also just to catch a potential bug. All models should have
            // an auto id so this would indicate an unexpected state.
            if (!this.isModel(model)) {
                throw new Error('Expected a model, plain object, or array of either');
            }

            // Make sure we don't add the same model twice.
            if (this.hasModelInRegistry(model)) {
                return;
            }

            // Add the model instance to this collection.
            this.models.push(model);
            this.onAdd(model);

            // We're assuming that the collection is not loading once a model is added.
            Vue.set(this, 'loading', false);

            return model;
        }

        /**
         * Called when a model has been removed from this collection.
         *
         * @param {Model} model
         */

    }, {
        key: 'onRemove',
        value: function onRemove(model) {
            model.unregisterCollection(this);
            this.removeModelFromRegistry(model);
            this.emit('remove', { model: model });
        }

        /**
         * Removes a model at a given index.
         *
         * @param  {number} index
          * @returns {Model} The model that was removed, or `undefined` if invalid.
         * @throws  {Error} If a model could not be found at the given index.
         */

    }, {
        key: '_removeModelAtIndex',
        value: function _removeModelAtIndex(index) {
            if (index < 0) {
                return;
            }

            var model = _.get(this.models, index);
            Vue.delete(this.models, index);
            this.onRemove(model);

            return model;
        }

        /**
         * Removes a `Model` from this collection.
         *
         * @param  {Model} model
         *
         * @return {Model}
         */

    }, {
        key: '_removeModel',
        value: function _removeModel(model) {
            return this._removeModelAtIndex(this.indexOf(model));
        }

        /**
         * Removes the given model from this collection.
         *
         * @param  {Model|Object|Array} model Model to remove, which can be a `Model`
         *                                    instance, an object to filter by,
         *                                    a function to filter by, or an array
         *                                    of any of the above to remove multiple.
         *
         * @return {Model|Model[]} The deleted model or an array of models if a filter
         *                         or array type was given.
         *
         * @throws {Error} If the model is an invalid type.
         */

    }, {
        key: 'remove',
        value: function remove(model) {
            if (!model) {
                throw new Error('Expected function, object, array, or model to remove');
            }

            // Support using a predicate to remove all models it returns true for.
            // Alternatively support an object of values to filter by.
            if (_.isFunction(model) || _.isPlainObject(model)) {
                return this.remove(_.filter(this.models, model));
            }

            // Support removing multiple models at the same time if an array was
            // given. A model would otherwise always be an object so this is safe.
            if (_.isArray(model)) {
                return _.filter(_.map(model, this.remove));
            }

            // This is just to catch a potential bug. All models should have
            // an auto id here so this would indicate an unexpected state.
            if (!this.isModel(model)) {
                throw new Error('Model to remove is not a valid model');
            }

            return this._removeModel(model);
        }

        /**
         * Determines whether a given value is an instance of a model.
         *
         * @param  {*} candidate A model candidate
         *
         * @return {boolean} `true` if the given `model` is an instance of Model.
         */

    }, {
        key: 'isModel',
        value: function isModel(candidate) {
            return _.isObject(candidate) && _.has(candidate, '_attributes') && _.has(candidate, '_uid');
        }

        /**
         * Returns the zero-based index of the given model in this collection.
         *
         * @see {@link https://lodash.com/docs/#findIndex}
         *
         * @return {number} the index of a model in this collection, or -1 if not found.
         */

    }, {
        key: 'indexOf',
        value: function indexOf(model) {
            var filter$$1 = model;

            // Getting the index of a model instance can be optimised.
            if (this.isModel(filter$$1)) {

                // Constant time check, if the registry doesn't have a record of
                // the model, we know it's not in the collection.
                if (!_.has(this._registry, model._uid)) {
                    return -1;
                }

                // There is no need to filter on the entire object, because the
                // unique ID of the model is all we need to identify it.
                filter$$1 = { _uid: model._uid };
            }

            return _.findIndex(this.models, filter$$1);
        }

        /**
         * @param {string|function|Object} where
         *
         * @return {Model} The first model that matches the given criteria, or
         *                 `undefined` if none could be found.
         *
         * @see {@link https://lodash.com/docs/#find}
         */

    }, {
        key: 'find',
        value: function find$$1(where) {
            return _.find(this.models, where);
        }

        /**
         * Wraps a new collection instance around some given models.
         */

    }, {
        key: 'wrap',
        value: function wrap(models) {
            return new this.constructor(models);
        }

        /**
         * Creates a new collection of the same type that contains only the models
         * for which the given predicate returns `true` for, or matches by property.
         *
         * @see {@link where}
         *
         * Important: Even though this returns a new collection, the references to
         *            each model are preserved, so changes will propagate to both.
         *
         * @param {function|Object|string} predicate Receives `model`.
         *
         * @returns {Collection}
         */

    }, {
        key: 'filter',
        value: function filter$$1(predicate) {
            return this.wrap(this.where(predicate));
        }

        /**
         * Returns the models for which the given predicate returns `true` for,
         * or models that match attributes in an object.
         *
         * @see {@link https://lodash.com/docs/#filter}
         *
         * @param {function|Object|string} predicate Receives `model`.
         *
         * @returns {Model[]}
         */

    }, {
        key: 'where',
        value: function where(predicate) {
            return _.filter(this.models, predicate);
        }

        /**
         * Returns an array that contains the returned result after applying a
         * function to each model in this collection.
         *
         * @see {@link https://lodash.com/docs/#map}
         *
         * @param {function} callback Receives `model`.
         *
         * @return {Model[]}
         */

    }, {
        key: 'map',
        value: function map$$1(callback) {
            return _.map(this.models, callback);
        }

        /**
         * Iterates through all models, calling a given callback for each one.
         *
         * @see {@link https://lodash.com/docs/#each}
         *
         * @param {function} callback Receives `model` and `index`.
         */

    }, {
        key: 'each',
        value: function each$$1(callback) {
            return _.each(this.models, callback);
        }

        /**
         * Reduces this collection to a value which is the accumulated result of
         * running each model through `iteratee`, where each successive invocation
         * is supplied the return value of the previous.
         *
         * If `initial` is not given, the first model of the collection is used
         * as the initial value.
         *
         * @param {function} iteratee Invoked with three arguments:
         *                            (result, model, index)
         *
         * @param {*} [initial] The initial value to use for the `result`.
         *
         * @returns {*} The final value of result, after the last iteration.
         */

    }, {
        key: 'reduce',
        value: function reduce$$1(iteratee, initial) {

            // Use the first model as the initial value if an initial was not given.
            if (arguments.length === 1) {
                initial = this.first();
            }

            return _.reduce(this.models, iteratee, initial);
        }

        /**
         * @param {function|string} iteratee Attribute name or callback to determine
         *                                   which values to sum by. Invoked with a
         *                                   single argument `model`.
         *
         * @returns {number} Sum of all models, accessed by attribute or callback.
         */

    }, {
        key: 'sum',
        value: function sum(iteratee) {
            return _.sumBy(this.models, iteratee);
        }

        /**
         * Returns an object composed of keys generated from the results of running
         * each model through `iteratee`. The corresponding value of each key is the
         * number of times the key was returned by iteratee.
         *
         * @see {@link https://lodash.com/docs/#countBy}
         *
         * @returns {Object}
         */

    }, {
        key: 'count',
        value: function count(iteratee) {
            return _.countBy(this.models, iteratee);
        }

        /**
         * Sorts this collection's models using a comparator. This method performs
         * a stable sort (it preserves the original sort order of equal elements).
         *
         * @see {@link https://lodash.com/docs/#sortBy}
         *
         * @param {function|string} comparator Attribute name or attribute function,
         *                                     invoked with a single arg `model`.
         */

    }, {
        key: 'sort',
        value: function sort(comparator) {
            Vue.set(this, 'models', _.sortBy(this.models, comparator));
        }

        /**
         * @param {Model|Object} model
         *
         * @returns {boolean} `true` if this collection contains the given model,
         *                    `false` otherwise.
         */

    }, {
        key: 'has',
        value: function has$$1(model) {
            return this.indexOf(model) >= 0;
        }

        /**
         * @returns {Model|undefined} The first model of this collection.
         */

    }, {
        key: 'first',
        value: function first$$1() {
            return _.first(this.models);
        }

        /**
         * @returns {Model|undefined} The last model of this collection.
         */

    }, {
        key: 'last',
        value: function last$$1() {
            return _.last(this.models);
        }

        /**
         * Removes and returns the first model of this collection, if there was one.
         *
         * @returns {Model|undefined} Removed model or undefined if there were none.
         */

    }, {
        key: 'shift',
        value: function shift() {
            if (!this.isEmpty()) {
                return this._removeModelAtIndex(0);
            }
        }

        /**
         * Removes and returns the last model of this collection, if there was one.
         *
         * @returns {Model|undefined} Removed model or undefined if there were none.
         */

    }, {
        key: 'pop',
        value: function pop() {
            if (!this.isEmpty()) {
                return this._removeModelAtIndex(this.size() - 1);
            }
        }

        /**
         * Replaces all models in this collection with those provided. This is
         * effectively equivalent to `clear` and `add`, and will result in an empty
         * collection if no models were provided.
         *
         * @param {Model|Model[]} models Models to replace the current models with.
         */

    }, {
        key: 'replace',
        value: function replace$$1(models) {
            this.clearModels();
            this.add(_.values(models));
        }

        /**
         * Returns the query parameters that should be used when paginating.
         *
         * @return {Object}
         */

    }, {
        key: 'getPaginationQuery',
        value: function getPaginationQuery() {
            return {
                page: this._page
            };
        }

        /**
         * @inheritDoc
         */

    }, {
        key: 'getFetchQuery',
        value: function getFetchQuery() {
            if (this.isPaginated()) {
                return this.getPaginationQuery();
            }

            return get$1(Collection.prototype.__proto__ || Object.getPrototypeOf(Collection.prototype), 'getFetchQuery', this).call(this);
        }

        /**
         * @param {Object} response
         *
         * @returns {Array|null} Models from the response.
         */

    }, {
        key: 'getModelsFromResponse',
        value: function getModelsFromResponse(response) {
            var models = response.getData();

            // An empty, non-array response indicates that we didn't intend to send
            // any models in the response. This means that the current models are
            // already up to date, as no changes are necessary.
            if (_.isNil(models) || models === '') {
                return null;
            }

            // We're making an assumption here that paginated models are returned
            // within the "data" field of the response.
            if (this.isPaginated()) {
                return _.get(models, 'data', models);
            }

            return models;
        }

        /**
         * Called when a save request was successful.
         *
         * @param {Object} response
         */

    }, {
        key: 'onSaveSuccess',
        value: function onSaveSuccess(response) {

            // Model data returned in the response.
            var saved = this.getModelsFromResponse(response);

            // All the models that are currently being saved.
            var saving = this.getSavingModels();

            // Empty response is similar to an empty response returned when saving
            // a model: assume that the attributes are the saved state, so sync.
            if (_.isNil(saved)) {
                _.each(saving, _.method('sync'));
            } else {

                // There is no sensible alternative to an array here, so anyting else
                // is considered an exception that indicates an unexpected state.
                if (!_.isArray(saved)) {
                    throw new ResponseError('Response data must be an array or empty', response);
                }

                // Check that the number of models returned in the response matches
                // the number of models that were saved. If these are not equal, it's
                // not possible to map saved data to the saving models.
                if (saved.length !== saving.length) {
                    throw new ResponseError('Expected the same number of models in the response', response);
                }

                // Update every model with its respective response data.
                // A strict requirement and assumption is that the models returned
                // in the response are in the same order as they are in the collection.
                _.each(saved, function (data, index) {
                    saving[index].onSaveSuccess(new ProxyResponse(200, data, response.getHeaders()));
                });
            }

            Vue.set(this, 'saving', false);
            Vue.set(this, 'fatal', false);

            this.emit('save', { error: null });
        }

        /**
         * @returns {Model[]} Models in this collection that are in a "saving" state.
         */

    }, {
        key: 'getSavingModels',
        value: function getSavingModels() {
            return _.filter(this.models, 'saving');
        }

        /**
         * @returns {Model[]} Models in this collection that are in a "deleting" state.
         */

    }, {
        key: 'getDeletingModels',
        value: function getDeletingModels() {
            return _.filter(this.models, 'deleting');
        }

        /**
         * Applies an array of validation errors to this collection's models.
         *
         * @param  {Array}   errors
         * @param  {integer} status Response status
         */

    }, {
        key: 'applyValidationErrorArray',
        value: function applyValidationErrorArray(errors) {
            var models = this.getSavingModels();

            // To allow matching errors with models, it's a strict requirement and
            // assumption that the array of errors returned in the response must have
            // the same number of elements as there are models being saved.
            if (errors.length !== models.length) {
                throw new ResponseError('Array of errors must equal the number of models');
            }

            // Set every model's errors in a way that emulates how saving a model
            // would fail in the same way.
            //
            // A strict requirement and assumption is that the models returned
            // in the response are in the same order as they are in the collection.
            _.each(models, function (model, index) {
                model.setErrors(errors[index]);
                Vue.set(model, 'saving', false);
                Vue.set(model, 'fatal', false);
            });
        }

        /**
         * Applies an object of validation errors keyed by model identifiers.
         *
         * @param  {Array}   errors
         * @param  {integer} status Response status
         */

    }, {
        key: 'applyValidationErrorObject',
        value: function applyValidationErrorObject(errors) {
            var lookup = _.keyBy(this.models, function (model) {
                return model.identifier();
            });

            _.each(errors, function (errors, identifier) {
                var model = _.get(lookup, identifier);

                if (model) {
                    model.setErrors(errors);
                }
            });
        }

        /**
         * Sets validation errors on this collection's models.
         *
         * @param {Array|Object} errors Either an array of length equal to the number
         *                              of models in this collection, or an object
         *                              of errors keyed by model identifiers.
         */

    }, {
        key: 'setErrors',
        value: function setErrors(errors) {

            // Support an array of errors, one for each model in the collection.
            if (_.isArray(errors)) {
                this.applyValidationErrorArray(errors);

                // Support an object of errors keyed by model identifiers.
            } else if (_.isPlainObject(errors)) {
                this.applyValidationErrorObject(errors);
            }
        }

        /**
         * @returns {Array} An array of this collection's validation errors.
         */

    }, {
        key: 'getErrors',
        value: function getErrors() {
            return _.map(this.models, 'errors');
        }

        /**
         * Called when a save request resulted in a validation error.
         *
         * @param {Object} response
         */

    }, {
        key: 'onSaveValidationFailure',
        value: function onSaveValidationFailure(error) {
            var response = error.getResponse();
            var errors = response.getValidationErrors();

            if (!_.isPlainObject(errors) && !_.isArray(errors)) {
                throw new ResponseError('Validation errors must be an object or array', response);
            }

            this.setErrors(errors);

            Vue.set(this, 'fatal', false);
            Vue.set(this, 'saving', false);
        }

        /**
         * Called when a save request resulted in an unexpected error,
         * eg. an internal server error (500)
         *
         * @param {Error}  error
         * @param {Object} response
         */

    }, {
        key: 'onFatalSaveFailure',
        value: function onFatalSaveFailure(error, response) {
            _.each(this.getSavingModels(), function (model) {
                model.onFatalSaveFailure(error, response);
            });

            Vue.set(this, 'fatal', true);
            Vue.set(this, 'saving', false);
        }

        /**
         * Called when a save request failed.
         *
         * @param {Error}  error
         * @param {Object} response
         */

    }, {
        key: 'onSaveFailure',
        value: function onSaveFailure(error) {
            if (this.isBackendValidationError(error)) {
                this.onSaveValidationFailure(error);

                // Not a validation error, so something else went wrong.
            } else {
                this.onFatalSaveFailure(error);
            }

            this.emit('save', { error: error });
        }

        /**
         * @returns {Array} The data to use for saving.
         */

    }, {
        key: 'getSaveData',
        value: function getSaveData() {
            return _.map(this.getSavingModels(), _.method('getSaveData'));
        }

        /**
         * Sets the page on this collection, enabling pagination. To disable
         * pagination on this collection, pass page as `null` or `undefined`.
         *
         * @param {number|boolean} [page] Page number, or `null` to disable.
         *
         * @returns {Collection} This collection.
         */

    }, {
        key: 'page',
        value: function page(_page) {

            // Disable pagination if a valid page wasn't provided.
            if (_.isNil(_page)) {
                Vue.set(this, '_page', NO_PAGE);

                // Page was provided, so we should either set the page or disable
                // pagination entirely if the page is `false`.
            } else {
                Vue.set(this, '_page', _.max([1, _.toSafeInteger(_page)]));
            }

            return this;
        }

        /**
         * @returns {integer|null} The page that this collection is on.
         */

    }, {
        key: 'getPage',
        value: function getPage() {
            return this._page;
        }

        /**
         * @returns {boolean} Whether this collection is currently paginated.
         */

    }, {
        key: 'isPaginated',
        value: function isPaginated() {
            return this._page !== NO_PAGE;
        }

        /**
         * @returns {boolean} Whether this collection is on the last page,
         *                            ie. there won't be more results that follow.
         */

    }, {
        key: 'isLastPage',
        value: function isLastPage() {
            return this._page === LAST_PAGE;
        }

        /**
         * Responsible for adjusting the page and appending of models that were
         * received by a paginated fetch request.
         *
         * @param {Model[]} models
         */

    }, {
        key: 'applyPagination',
        value: function applyPagination(models) {

            // If no models were returned in the response we can assume that
            // we're now on the last page, and we should not continue.
            if (_.isEmpty(models)) {
                Vue.set(this, '_page', LAST_PAGE);

                // Otherwise, there were at least one model, and we can safely
                // assume that we want to increment the page number.
            } else {
                Vue.set(this, '_page', this._page + 1);
                this.add(models);
            }
        }

        /**
         * Called when a fetch request was successful.
         *
         * @param {Object} response
         */

    }, {
        key: 'onFetchSuccess',
        value: function onFetchSuccess(response) {
            var models = this.getModelsFromResponse(response);

            // There is no sensible alternative to an array here, so anyting else
            // is considered an exception that indicates an unexpected state.
            if (!_.isArray(models)) {
                throw new ResponseError('Expected an array of models in fetch response');
            }

            // Append via pagination.
            if (this.isPaginated()) {
                this.applyPagination(models);

                // Replace all current models with the fetched ones.
            } else {
                this.replace(models);
            }

            Vue.set(this, 'loading', false);
            Vue.set(this, 'fatal', false);

            this.emit('fetch', { error: null });
        }

        /**
         * Called when a fetch request failed.
         *
         * @param {Error}  error
         */

    }, {
        key: 'onFetchFailure',
        value: function onFetchFailure(error) {
            this.clearErrors();

            Vue.set(this, 'fatal', true);
            Vue.set(this, 'loading', false);

            this.emit('fetch', { error: error });
        }

        /**
         * Called before a fetch request is made.
         *
         * @returns {boolean|undefined} `false` if the request should not be made.
         */

    }, {
        key: 'onFetch',
        value: function onFetch() {

            // Don't fetch if there are no more results to be fetched.
            if (this.isPaginated() && this.isLastPage()) {
                return false;
            }

            // Because we're fetching new data, we can assume that this collection
            // is now loading. This allows the template to indicate a loading state.
            Vue.set(this, 'loading', true);
        }

        /**
         * Called when a delete request was successful.
         *
         * @param {Object} response
         */

    }, {
        key: 'onDeleteSuccess',
        value: function onDeleteSuccess(response) {
            Vue.set(this, 'deleting', false);
            Vue.set(this, 'fatal', false);

            _.each(this.getDeletingModels(), function (model) {
                model.onDeleteSuccess(response);
            });

            this.emit('delete', { error: null });
        }

        /**
         * Called when a delete request resulted in a general error.
         *
         * @param {Error}  error
         * @param {Object} response
         */

    }, {
        key: 'onDeleteFailure',
        value: function onDeleteFailure(error) {
            Vue.set(this, 'fatal', true);
            Vue.set(this, 'deleting', false);

            _.each(this.getDeletingModels(), function (model) {
                model.onDeleteFailure(error);
            });

            this.emit('delete', { error: error });
        }

        /**
         * Called before a save request is made.
         *
         * @returns {boolean} Either `true` or false` if the request should not be
         *                    made, where `true` indicates that the request should
         *                    be considered a "success" rather than a "cancel".
         *
         */

    }, {
        key: 'onSave',
        value: function onSave() {

            // Don't save if we're already busy saving this collection.
            // This prevents things like accidental double clicks.
            if (this.saving) {
                return false;
            }

            var valid = true;

            // Call 'onSave' on each model so that the models can set their state
            // accordingly, and indicate whether a validation failure should occur.
            _.each(this.models, function (model) {
                try {
                    model.onSave();
                } catch (error) {
                    if (error instanceof ValidationError) {
                        valid = false;
                    } else {
                        throw error;
                    }
                }
            });

            // Throwing a validation error here will cause the save request to be
            // rejected, because at least one model's data is not valid.
            if (!valid) {
                throw new ValidationError(this.getErrors());
            }

            Vue.set(this, 'saving', true);
        }

        /**
         * Collect all model identifiers.
         *
         * @returns {Array}
         */

    }, {
        key: 'getIdentifiers',
        value: function getIdentifiers(models) {
            return _.map(models, _.method('identifier'));
        }

        /**
         * @inheritDoc
         */

    }, {
        key: 'getDeleteBody',
        value: function getDeleteBody() {
            if (this.getOption('useDeleteBody')) {
                return this.getIdentifiers(this.getDeletingModels());
            }

            return {};
        }

        /**
         * @returns {string} The query parameter key to use for model identifiers.
         */

    }, {
        key: 'getDeleteQueryIdenitifierKey',
        value: function getDeleteQueryIdenitifierKey() {
            return 'id';
        }

        /**
         * @inheritDoc
         */

    }, {
        key: 'getDeleteQuery',
        value: function getDeleteQuery() {

            // Don't use query parameters if we want send the request data in the body.
            if (this.getOption('useDeleteBody')) {
                return {};
            }

            // Collect all the identifiers of the models being deleted.
            var models = this.getDeletingModels();
            var identifier = this.getDeleteQueryIdenitifierKey();
            var identifiers = this.getIdentifiers(models);

            return defineProperty({}, identifier, _.join(identifiers, ','));
        }

        /**
         * Called before a delete request is made.
         *
         * @returns {boolean} `false` if the request should not be made.
         */

    }, {
        key: 'onDelete',
        value: function onDelete() {

            // Don't save if we're already busy saving this collection.
            // This prevents things like accidental double clicks.
            if (this.deleting) {
                return false;
            }

            // Exclude all models that return `false` on delete.
            var models = _.filter(this.models, function (model) {
                return model.onDelete() !== false;
            });

            // Don't save if there are no models to delete.
            if (_.isEmpty(models)) {
                return true;
            }

            Vue.set(this, 'deleting', true);
        }
    }]);
    return Collection;
}(Base);

/**
 * Reserved keywords that can't be used for attribute or option names.
 */
var RESERVED = _.invert(['_attributes', '_collections', '_errors', '_listeners', '_reference', '_registry', '_uid', 'attributes', 'collections', 'deleting', 'errors', 'fatal', 'loading', 'memoized', 'models', 'saving']);

/**
 * Recursive deep copy helper that honours the "clone" function of models and
 * collections. This is required to support nested instances.
 *
 * @param {Object} source
 * @param {Object} target
 * @param {Array}  keys     Optional
 */
var copyFrom = function copyFrom(source, target, keys$$1) {
    if (keys$$1) {
        source = _.pick(source, keys$$1);
    }

    _.each(source, function (value, key) {
        if (_.isArray(value)) {
            Vue.set(target, key, []);
            copyFrom(value, target[key]);
        } else if (_.isPlainObject(value)) {
            Vue.set(target, key, {});
            copyFrom(value, target[key]);
        } else if (_.isObject(value) && _.isFunction(value.clone)) {
            Vue.set(target, key, value.clone());
        } else {
            Vue.set(target, key, _.cloneDeep(value));
        }
    });
};

/**
 * Base model class.
 */

var Model = function (_Base) {
    inherits(Model, _Base);
    createClass(Model, [{
        key: '$',


        /**
         * A convenience wrapper around the model's attributes that are saved.
         * This is similar to the `saved` method, but instead of accessing a single
         * property it returns the whole saved object, so that you can do something
         * like model.$.attribute when you want to display it somewhere.
         *
         * @returns {Object} This model's saved, reference data.
         */
        get: function get$$1() {
            return this._reference;
        }

        /**
         * @returns {Object} This model's "active" state attributes.
         */

    }, {
        key: 'attributes',
        get: function get$$1() {
            return this._attributes;
        }

        /**
         * @returns {Object} The collection that this model is registered to.
         */

    }, {
        key: 'collections',
        get: function get$$1() {
            return _.values(this._collections);
        }

        /**
         * @returns {Object} This model's errors, which are cleared automatically.
         */

    }, {
        key: 'errors',
        get: function get$$1() {
            return this.getErrors();
        }

        /**
         * Creates a new instance, called when using 'new'.
         *
         * @param  {Object}     [attributes]  Model attributes
         * @param  {Collection} [collection]  Collection that this model belongs to.
         * @param  {Object}     [options]     Options to set on the model.
         */

    }]);

    function Model() {
        var attributes = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
        var collection = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
        var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
        classCallCheck(this, Model);

        var _this = possibleConstructorReturn(this, (Model.__proto__ || Object.getPrototypeOf(Model)).call(this, options));

        Vue.set(_this, '_collections', {}); // Collections that contain this model.
        Vue.set(_this, '_reference', {}); // Saved attribute state.
        Vue.set(_this, '_attributes', {}); // Active attribute state.
        Vue.set(_this, '_mutations', {}); // Mutator cache.
        Vue.set(_this, '_errors', {}); // Validation errors.

        _this.clearState();

        // Cache certain methods that don't need to be evaluated more than once.
        _this.memoize();

        // Cache mutator pipelines so that they can run as a single function.
        _this.compileMutators();

        // Assign all given model data to the model's attributes and reference.
        _this.assign(attributes);

        // Register the given collection (if any) to the model. This is so that
        // the model can be added to the collection automatically when it is
        // created on save, or removed on delete.
        if (collection) {
            _this.registerCollection(collection);
        }
        return _this;
    }

    /**
     * Creates a copy of this model, with the same attributes and options. The
     * clone will also belong to the same collections as the subject.
     *
     * @returns {Model}
     */


    createClass(Model, [{
        key: 'clone',
        value: function clone() {
            var attributes = {};
            var reference = {};

            // Clone all attributes and their descriptors.
            copyFrom(this._attributes, attributes);
            copyFrom(this._reference, reference);

            // Create a copy.
            var clone = new this.constructor();

            // Make sure that the clone belongs to the same collections.
            clone.registerCollection(this._collections);

            // Make sure that the clone has the same existing options.
            clone.setOptions(this.getOptions());

            Vue.set(clone, '_reference', reference);
            Vue.set(clone, '_attributes', attributes);

            return clone;
        }

        /**
         * Prepare certain methods to only be called once. These are methods that
         * are expected to return the same data every time.
         *
         * @see {@link https://lodash.com/docs/#once}
         */

    }, {
        key: 'memoize',
        value: function memoize() {
            var _this2 = this;

            var memoized = ['validation', //   \
            'defaults', //   | These do not need to be evaluated every time.
            'routes'];

            _.each(memoized, function (name) {
                return _this2[name] = _.once(_this2[name]);
            });
        }

        /**
         * Returns the model's identifier value.
         */

    }, {
        key: 'identifier',
        value: function identifier() {
            return this.get(this.getOption('identifier'));
        }

        /**
         * @returns {Object} An empty representation of this model.
         *                   It's important that all model attributes have a default
         *                   value in order to be reactive in Vue.
         */

    }, {
        key: 'defaults',
        value: function defaults$$1() {
            return {};
        }

        /**
         * @returns {Object} Attribute mutations keyed by attribute name.
         */

    }, {
        key: 'mutations',
        value: function mutations() {
            return {};
        }

        /**
         * Add validation rules here, or use option?
         */

    }, {
        key: 'validation',
        value: function validation() {
            return {};
        }

        /**
         * Returns the default options for this model.
         *
         * @returns {Object}
         */

    }, {
        key: 'getDefaultOptions',
        value: function getDefaultOptions() {
            return _.merge({}, get$1(Model.prototype.__proto__ || Object.getPrototypeOf(Model.prototype), 'getDefaultOptions', this).call(this), {

                // The attribute that should be used to uniquely identify this model.
                identifier: 'id',

                // Whether this model should allow an existing identifier to be
                // overwritten on update.
                overwriteIdentifier: false,

                // Whether this model should perform a "patch" on update,
                // which will only send changed attributes in the request.
                patch: false,

                // Whether this model should save even if no attributes have changed
                // since the last time they were synced. If set to `false` and no
                // changes have been made, the request will be considered a success.
                saveUnchanged: true,

                // Whether this model should only use the first validation error it
                // receives, rather than an array of errors.
                useFirstErrorOnly: false,

                // Whether this model should validate an attribute that has changed.
                // This would only affect the errors of the changed attribute and
                // will only be applied if the value is not a blank string.
                validateOnChange: false,

                // Whether this model should validate models and collections within
                // its attribute tree. The result is implicit recursion as each of
                // those instances will also validate their trees, etc.
                validateRecursively: true,

                // Whether this model should mutate a property as it is changed,
                // before it is set. This is a rare requirement because you usually
                // don't  want to mutate something that you are busy editing.
                mutateOnChange: false,

                // Whether this model should mutate all attributes before they are
                // synced to the "saved" state. This would include construction,
                // on fetch, on save, and on assign.
                mutateBeforeSync: true,

                // Whether this model should use mutated values for the attributes
                // in "save" request. This will not mutate the active state.
                mutateBeforeSave: true
            });
        }

        /**
         * Compiles all mutations into pipelines that can be executed quickly.
         */

    }, {
        key: 'compileMutators',
        value: function compileMutators() {
            this._mutations = _.mapValues(this.mutations(), function (m) {
                return _.flow(m);
            });
        }

        /**
         * @returns {Object} Parameters to use for replacement in route patterns.
         */

    }, {
        key: 'getRouteParameters',
        value: function getRouteParameters() {
            return _.merge({}, get$1(Model.prototype.__proto__ || Object.getPrototypeOf(Model.prototype), 'getRouteParameters', this).call(this), this._attributes);
        }

        /**
         * Registers a collection on this model. When this model is created it will
         * automatically be added to the collection. Similarly, when this model is
         * delete it will be remove from the collection. Registering the same
         * collection more than once has no effect.
         *
         * @param {Collection} collection
         */

    }, {
        key: 'registerCollection',
        value: function registerCollection(collection) {
            if (_.isArray(collection)) {
                _.each(collection, this.registerCollection);
                return;
            }

            if (_.isNil(collection) || _.isUndefined(collection._uid)) {
                throw new Error('Collection is not valid');
            }

            Vue.set(this._collections, collection._uid, collection);
        }

        /**
         * Removes a collection from this model's collection registry, removing all
         * effects that would occur when creating or deleting this model.
         *
         * Unregistering a collection that isn't registered has no effect.
         *
         * @param {Collection} collection
         */

    }, {
        key: 'unregisterCollection',
        value: function unregisterCollection(collection) {
            if (_.isArray(collection)) {
                _.each(collection, this.unregisterCollection);
                return;
            }

            if (_.isNil(collection) || _.isUndefined(collection._uid)) {
                throw new Error('Collection is not valid');
            }

            Vue.delete(this._collections, collection._uid);
        }

        /**
         * Reverts all attributes back to their defaults, and completely removes all
         * attributes that don't have defaults. This will also sync the reference
         * attributes, and is not reversable.
         */

    }, {
        key: 'clearAttributes',
        value: function clearAttributes() {
            var defaults$$1 = this.defaults();

            Vue.set(this, '_attributes', _.cloneDeep(defaults$$1));
            Vue.set(this, '_reference', _.cloneDeep(defaults$$1));
        }

        /**
         * Reverts all attributes back to their defaults, and completely removes all
         * attributes that don't have defaults. This will also sync the reference
         * attributes, and is not reversable.
         */

    }, {
        key: 'clear',
        value: function clear() {
            this.clearAttributes();
            this.clearErrors();
            this.clearState();
        }

        /**
         * Resets model state, ie. `loading`, etc back to their initial states.
         */

    }, {
        key: 'clearState',
        value: function clearState() {
            Vue.set(this, 'loading', false);
            Vue.set(this, 'saving', false);
            Vue.set(this, 'deleting', false);
            Vue.set(this, 'fatal', false);
        }

        /**
         * Assigns all given model data to the model's attributes and reference.
         * This will also fill any gaps with the model's default attribute.
         *
         * @param {Object} attributes
         *
         * @returns {Object} The attributes that were assigned to the model.
         */

    }, {
        key: 'assign',
        value: function assign(attributes) {
            this.set(_.defaultsDeep({}, attributes, _.cloneDeep(this.defaults())));
            this.sync();
        }

        /**
         * Resets all attributes back to their reference values (source of truth).
         * A good use case for this is when form fields are bound directly to the
         * model's attributes. Changing values in the form fields will change the
         * attributes on the model. On cancel, you can revert the model back to
         * its saved, original state using reset().
         *
         * You can also pass one or an array of attributes to reset.
         *
         * @param {string|string[]} attribute
         */

    }, {
        key: 'reset',
        value: function reset(attribute) {

            // Reset specific attributes.
            if (attribute) {
                copyFrom(this._reference, this._attributes, _.castArray(attribute));

                // Reset all attributes if one or more specific ones were not given.
            } else {
                copyFrom(this._reference, this._attributes);
            }

            this.clearErrors();
            this.emit('reset');
        }

        /**
         * @returns {*} The value of an attribute after applying its mutations.
         */

    }, {
        key: 'mutated',
        value: function mutated(attribute, value) {
            var mutator = _.get(this._mutations, attribute);

            if (mutator) {
                return mutator(value);
            }

            return value;
        }

        /**
         * Mutates either specific attributes or all attributes if none provided.
         * @param {string|string[]|undefined} attribute
         */

    }, {
        key: 'mutate',
        value: function mutate(attribute) {
            var _this3 = this;

            if (_.isUndefined(attribute)) {
                _.each(this._attributes, function (value, attribute) {
                    Vue.set(_this3._attributes, attribute, _this3.mutated(attribute, value));
                });

                // Only mutate specific attributes.
            } else {
                _.each(_.castArray(attribute), function (attribute) {
                    var current = _this3.get(attribute);
                    var mutated = _this3.mutated(attribute, current);

                    Vue.set(_this3._attributes, attribute, mutated);
                });
            }
        }

        /**
         * Sync the current attributes to the reference attributes. This is usually
         * only called on save. We have to clone the values otherwise we
         * end up with references to the same object in both attribute sets.
         *
         * You can also pass one or an array of attributes to sync.
         *
         * @param {string|string[]} attribute
         */

    }, {
        key: 'sync',
        value: function sync(attribute) {
            var _this4 = this;

            // Mutate all attributes before we sync them, if required to do so.
            if (this.getOption('mutateBeforeSync')) {
                this.mutate(attribute);
            }

            // We're cloning deep to avoid multiple references to the same object,
            // otherwise updating the attributes will also update the reference.
            // Set each saved attribute to its active equivalent.
            var active = _.cloneDeep(this._attributes);

            // Sync either specific attributes or all attributes if none provided.
            if (_.isUndefined(attribute)) {
                Vue.set(this, '_reference', active);
            } else {
                _.each(_.castArray(attribute), function (attribute) {
                    Vue.set(_this4._reference, attribute, _.get(active, attribute));
                });
            }

            this.emit('sync');
        }

        /**
         * Registers an attribute on this model so that it can be accessed directly
         * on the model, passing through `get` and `set`.
         */

    }, {
        key: 'registerAttribute',
        value: function registerAttribute(attribute) {
            var _this5 = this;

            // Protect against unwillingly using an attribute name that already
            // exists as an internal property or method name.
            if (_.has(RESERVED, attribute)) {
                throw new Error('Can\'t use reserved attribute name \'' + attribute + '\'');
            }

            // Create dynamic accessors and mutations so that we can update the
            // model directly while also keeping the model attributes in sync.
            Object.defineProperty(this, attribute, {
                get: function get$$1() {
                    return _this5.get(attribute);
                },
                set: function set$$1(value) {
                    return _this5.set(attribute, value);
                }
            });
        }

        /**
         * Sets the value of an attribute and registers the magic "getter" in a way
         * that is compatible with Vue's reactivity. This method should always be
         * used when setting the value of an attribute.
         *
         * @param  {string|Object}  attribute
         * @param  {*}              value
         *
         * @returns {*} The value that was set.
         */

    }, {
        key: 'set',
        value: function set$$1(attribute, value) {
            var _this6 = this;

            // Allow batch set of multiple attributes at once, ie. set({...});
            if (_.isPlainObject(attribute)) {
                _.each(attribute, function (value, key) {
                    _this6.set(key, value);
                });

                return;
            }

            var defined = this.has(attribute);

            // Only register the pass-through property if it's not already set up.
            // If it already exists on the instance, we know it has been.
            if (!defined) {
                this.registerAttribute(attribute);
            }

            // Current value of the attribute, or `undefined` if not set
            var previous = this.get(attribute);

            // Run the attribute's mutations if required to do so on change.
            if (this.getOption('mutateOnChange')) {
                value = this.mutated(attribute, value);
            }

            Vue.set(this._attributes, attribute, value);

            // Only consider a change if the attribute was already defined.
            var changed = defined && !_.isEqual(previous, value);

            if (changed) {
                this.emit('change', { attribute: attribute, previous: previous, value: value });

                // Validate on change only if it's not the first time it's set.
                if (this.getOption('validateOnChange')) {
                    Vue.nextTick(function () {
                        return _this6.validateAttribute(attribute);
                    });
                }
            }

            return value;
        }

        /**
         * Reverts all attributes back to their defaults, or `undefined` if a
         * default value is not defined.
         *
         * You can also pass one or an array of attributes to unset.
         *
         * @param {string|string[]} attribute
         */

    }, {
        key: 'unset',
        value: function unset$$1(attribute) {
            var _this7 = this;

            // We're cloning deep to avoid multiple references to the same object,
            // otherwise updating the attributes will also update the reference.
            var defaults$$1 = _.cloneDeep(this.defaults());

            // Unset either specific attributes or all attributes if none provided.
            var attributes = _.defaultTo(attribute, _.keys(this._attributes));

            // Unset either specific attributes or all attributes if none provided.
            _.each(_.castArray(attributes), function (attribute) {
                if (_this7.has(attribute)) {
                    Vue.set(_this7._attributes, attribute, _.get(defaults$$1, attribute));
                }
            });
        }

        /**
         * Similar to `saved`, returns an attribute's value or a fallback value
         * if this model doesn't have the attribute.
         *
         * @param {string} attribute
         * @param {*}      fallback
         *
         * @returns {*} The value of the attribute or `fallback` if not found.
         */

    }, {
        key: 'get',
        value: function get$$1(attribute, fallback) {
            return _.get(this._attributes, attribute, fallback);
        }

        /**
         * Similar to `get`, but accesses the saved attributes instead.
         *
         * This is useful in cases where you want to display an attribute but also
         * change it. For example, a modal with a title based on a model field, but
         * you're also editing that field. The title will be updating reactively if
         * it's bound to the active attribute, so bind to the saved one instead.
         *
         * @param {string} attribute
         * @param {*}      fallback
         *
         * @returns {*} The value of the attribute or `fallback` if not found.
         */

    }, {
        key: 'saved',
        value: function saved(attribute, fallback) {
            return _.get(this._reference, attribute, fallback);
        }

        /**
         * Determines if the model has an attribute.
         *
         * @param  {string}  attribute
         * @returns {boolean} `true` if an attribute exists, `false` otherwise.
         *                   Will return true if the object exists but is undefined.
         */

    }, {
        key: 'has',
        value: function has$$1(attribute) {
            return _.has(this._attributes, attribute);
        }

        /**
         * Validates a specific attribute of this model, and sets errors for it.
         *
         * @returns {boolean} `true` if valid, `false` otherwise.
         */

    }, {
        key: 'validateAttribute',
        value: function validateAttribute(attribute) {
            var _this8 = this;

            var value = this.get(attribute);
            var rules = this.validation();
            var valid = true;
            var errors = [];

            if (attribute in rules) {
                var ruleset = _.castArray(rules[attribute]);

                _.each(ruleset, function (rule) {
                    var result = rule(value, attribute, _this8);

                    // Rules should return an error message if validation failed.
                    if (_.isString(result)) {
                        errors.push(result);
                        valid = false;

                        // Break early if we're only interested in the first error.
                        if (_this8.getOption('useFirstErrorOnly')) {
                            return false;
                        }
                    }
                });
            }

            // Defer validation if an attribute is an object that has a `validate`
            // method. The expectation is that the validate function will return
            // `true` if valid, `false` if not, and handle its own errors.
            if (this.getOption('validateRecursively')) {
                if (_.isFunction(_.get(value, 'validate'))) {
                    valid = value.validate() && valid;
                }
            }

            // Set the errors for the attribute.
            this.setAttributeErrors(attribute, errors);

            return valid;
        }

        /**
         * Validates all attributes.
         *
         * @param {Object} [attributes] One or more attributes to validate.
         *
         * @returns {boolean} `true` if the model passes validation.
         */

    }, {
        key: 'validate',
        value: function validate(attributes) {
            var _this9 = this;

            if (_.isString(attributes)) {
                return this.validateAttribute(attributes);

                // Only validate the attributes that were specified.
            } else if (_.isArray(attributes)) {
                attributes = _.pick(this._attributes, attributes);

                // Or validate all attributes if none were given.
            } else if (_.isUndefined(attributes)) {
                attributes = this._attributes;
            } else {
                throw new Error('Validation attributes must be an array, a string, or not given');
            }

            // Validate all attributes if none were given.
            return _.reduce(attributes, function (valid, value, attribute) {
                return _this9.validateAttribute(attribute) && valid;
            }, true);
        }

        /**
         * @returns {Object} A native representation of this model that will determine
         *                   the contents of JSON.stringify(model).
         */

    }, {
        key: 'toJSON',
        value: function toJSON() {
            return this._attributes;
        }

        /**
         * Adds this model to all registered collections.
         */

    }, {
        key: 'addToAllCollections',
        value: function addToAllCollections() {
            var _this10 = this;

            _.each(this._collections, function (collection, id) {
                collection.add(_this10);
            });
        }

        /**
         * Removes this model from all registered collections.
         */

    }, {
        key: 'removeFromAllCollections',
        value: function removeFromAllCollections() {
            var _this11 = this;

            _.each(this._collections, function (collection, id) {
                collection.remove(_this11);
            });
        }

        /**
         * Returns an array of attribute names that have changed, or `false` if no
         * changes have been made since the last time this model was synced.
         *
         * @returns {Array|boolean} An array of changed attribute names, or `false`
         *                         if no attributes have changed since the last sync.
         */

    }, {
        key: 'changed',
        value: function changed() {
            var _this12 = this;

            var changed = [];

            _.each(this._attributes, function (value, attribute) {
                if (!_.isEqual(value, _this12.saved(attribute))) {
                    changed.push(attribute);
                }
            });

            return !_.isEmpty(changed) ? changed : false;
        }

        /**
         * Called when a fetch request was successful.
         */

    }, {
        key: 'onFetchSuccess',
        value: function onFetchSuccess(response) {
            var attributes = response.getData();

            // A fetch request must receive *some* data in return.
            if (_.isEmpty(attributes)) {
                throw new ResponseError("No data in fetch response", response);
            }

            this.assign(attributes);

            Vue.set(this, 'fatal', false);
            Vue.set(this, 'loading', false);

            this.emit('fetch', { error: null });
        }

        /**
         * Called when a fetch request failed.
         *
         * @param {Error}  error
         */

    }, {
        key: 'onFetchFailure',
        value: function onFetchFailure(error) {
            Vue.set(this, 'fatal', true);
            Vue.set(this, 'loading', false);

            this.emit('fetch', { error: error });
        }

        /**
         * @returns {string} The key to use when generating the `patch` URL.
         */

    }, {
        key: 'getPatchRoute',
        value: function getPatchRoute() {
            return this.getRoute('patch', 'save');
        }

        /**
         * @returns {string} The key to use when generating the `create` URL.
         */

    }, {
        key: 'getCreateRoute',
        value: function getCreateRoute() {
            return this.getRoute('create', 'save');
        }

        /**
         * @returns {string} The key to use when generating the `update` URL.
         */

    }, {
        key: 'getUpdateRoute',
        value: function getUpdateRoute() {
            if (this.shouldPatch()) {
                return this.getPatchRoute();
            }

            return this.getRoute('update', 'save');
        }

        /**
         * @returns {string} The method to use when making an update request.
         */

    }, {
        key: 'getUpdateMethod',
        value: function getUpdateMethod() {
            return this.shouldPatch() ? this.getPatchMethod() : get$1(Model.prototype.__proto__ || Object.getPrototypeOf(Model.prototype), 'getUpdateMethod', this).call(this);
        }

        /**
         * @returns {string} The method to use when making an save request.
         */

    }, {
        key: 'getSaveMethod',
        value: function getSaveMethod() {
            return this.isNew() ? this.getCreateMethod() : this.getUpdateMethod();
        }

        /**
         * @inheritDoc
         */

    }, {
        key: 'getSaveRoute',
        value: function getSaveRoute() {
            if (this.isNew()) {
                return this.getCreateRoute();
            }

            return this.getUpdateRoute();
        }

        /**
         * Returns whether this model should perform a "patch" on update, which will
         * only send changed data in the request, rather than all attributes.
         *
         * @returns {boolean} Whether this model should perform a "patch" on update,
         *                    which will only send changed data in the request,
         *                    rather than all attributes.
         */

    }, {
        key: 'shouldPatch',
        value: function shouldPatch() {
            return Boolean(this.getOption('patch'));
        }

        /**
         * @returns {Object} The data to send to the server when saving this model.
         */

    }, {
        key: 'getSaveData',
        value: function getSaveData() {

            // Only use changed attributes if patching.
            if (this.isExisting() && this.shouldPatch()) {
                return _.pick(this._attributes, this.changed());
            }

            return this._attributes;
        }

        /**
         * @returns {*} A potential identifier parsed from response data.
         */

    }, {
        key: 'parseIdentifier',
        value: function parseIdentifier(data) {
            return data;
        }

        /**
         * @returns {boolean} Whether the given identifier is considered a valid
         *                   identifier value for this model.
         */

    }, {
        key: 'isValidIdentifier',
        value: function isValidIdentifier(identifier) {
            return Boolean(identifier);
        }

        /**
         * @returns {boolean} Whether this model allows an existing identifier to be
         *                    overwritten on update.
         */

    }, {
        key: 'shouldAllowIdentifierOverwrite',
        value: function shouldAllowIdentifierOverwrite() {
            return Boolean(this.getOption('overwriteIdentifier'));
        }

        /**
         * Updates the model data with data returned from the server.
         *
         * @param {Object} response
         */

    }, {
        key: 'update',
        value: function update(data) {

            // No content means we don't want to update the model at all.
            // The attributes that we passed in the request should now be considered
            // the source of truth, so we should update the reference attributes here.
            if (!data || _.isObjectLike(data) && _.isEmpty(data)) {
                this.sync();

                // A plain object implies that we want to update the model data.
                // It's not a requirement to respond with a complete dataset,
                // eg. a response to a patch request might return partial data.
            } else if (_.isPlainObject(data)) {
                this.assign(data);

                // There is some data, but it's not an object, so we can assume that the
                // response only returned an identifier for this model.
            } else {
                var identifier = this.parseIdentifier(data);

                // It's possible that the response didn't actually return a valid
                // identifier, so before we try to use it we should make sure that
                // we're not accidentially assigning the wrong data as identifiers.
                if (this.isValidIdentifier(identifier)) {

                    // The current identifier of this model.
                    var current = this.identifier();

                    // If an identifier already exists on this model and the returned
                    // identifier is not the same, this almost definitely indicates
                    // an unexpected state. The default is to protect against this
                    // and fail hard, but this might not always be what we want.
                    if (current && identifier !== current) {
                        if (!this.shouldAllowIdentifierOverwrite()) {
                            throw new Error('Not allowed to overwrite model identifier');
                        }
                    }

                    // Update the identifier and sync the saved data.
                    this.set(this.getOption('identifier'), identifier);
                    this.sync();
                } else {
                    throw new Error('Expected an empty response, object, or valid identifier');
                }
            }
        }

        /**
         * Sets errors for a specific attribute. Support the ability to clear error
         * by passing an empty value.
         *
         * @param {string}       attribute
         * @param {string|array} errors
         */

    }, {
        key: 'setAttributeErrors',
        value: function setAttributeErrors(attribute, errors) {
            if (_.isEmpty(errors)) {
                Vue.delete(this._errors, attribute);
            } else {
                Vue.set(this._errors, attribute, _.castArray(errors));
            }
        }

        /**
         * Sets the errors on this model.
         *
         * @param {Object} errors
         */

    }, {
        key: 'setErrors',
        value: function setErrors(errors) {
            var _this13 = this;

            if (_.isEmpty(errors)) {
                Vue.set(this, '_errors', {});
                return;
            }

            _.each(errors, function (errors, attribute) {
                _this13.setAttributeErrors(attribute, errors);
            });
        }

        /**
         * @returns {Object} Validation errors on this model.
         */

    }, {
        key: 'getErrors',
        value: function getErrors() {
            if (this.getOption('useFirstErrorOnly')) {
                return _.mapValues(this._errors, _.head);
            }

            return this._errors;
        }

        /**
         * Clears all errors on this model.
         */

    }, {
        key: 'clearErrors',
        value: function clearErrors() {
            this.setErrors({});
            Vue.set(this, 'fatal', false);
        }

        /**
         * Called when a save request was successful.
         *
         * @param {Object|null} response
         */

    }, {
        key: 'onSaveSuccess',
        value: function onSaveSuccess(response) {

            // Clear errors because the request was successful.
            this.clearErrors();

            // Update this model with the data that was returned in the response.
            if (response) {
                this.update(response.getData());
            }

            Vue.set(this, 'saving', false);
            Vue.set(this, 'fatal', false);

            // Automatically add to all registered collections.
            this.addToAllCollections();

            this.emit('save', { error: null });
        }

        /**
         * Called when a save request resulted in a validation error.
         *
         * @param {Object} errors
         */

    }, {
        key: 'onSaveValidationFailure',
        value: function onSaveValidationFailure(error) {
            var errors = error.getResponse().getValidationErrors();

            if (!_.isPlainObject(errors)) {
                throw new ResponseError('Validation errors must be an object', error.getResponse());
            }

            this.setErrors(errors);

            Vue.set(this, 'fatal', false);
            Vue.set(this, 'saving', false);
        }

        /**
         * Called when a save request resulted in an unexpected error,
         * eg. an internal server error (500)
         *
         * @param {Error}  error
         * @param {Object} response
         */

    }, {
        key: 'onFatalSaveFailure',
        value: function onFatalSaveFailure(error) {
            this.clearErrors();

            Vue.set(this, 'fatal', true);
            Vue.set(this, 'saving', false);
        }

        /**
         * Called when a save request resulted in a general error.
         *
         * @param {Error}  error
         * @param {Object} response
         */

    }, {
        key: 'onSaveFailure',
        value: function onSaveFailure(error) {
            if (this.isBackendValidationError(error)) {
                this.onSaveValidationFailure(error);
            } else {
                this.onFatalSaveFailure(error);
            }

            this.emit('save', { error: error });
        }

        /**
         * Called when a delete request was successful.
         */

    }, {
        key: 'onDeleteSuccess',
        value: function onDeleteSuccess(response) {
            this.clear();
            this.removeFromAllCollections();

            Vue.set(this, 'deleting', false);
            Vue.set(this, 'fatal', false);

            this.emit('delete', { error: null });
        }

        /**
         * Called when a delete request resulted in a general error.
         *
         * @param {Error}  error
         */

    }, {
        key: 'onDeleteFailure',
        value: function onDeleteFailure(error) {
            Vue.set(this, 'deleting', false);
            Vue.set(this, 'fatal', true);

            this.emit('delete', { error: error });
        }

        /**
         * Called before a fetch request is made.
         *
         * @returns {boolean|undefined} `false` if the request should not be made.
         */

    }, {
        key: 'onFetch',
        value: function onFetch() {

            // Don't fetch if already fetching. This prevents accidental requests
            // that sometimes occur as a result of a double-click.
            if (this.loading) {
                return false;
            }

            Vue.set(this, 'loading', true);
        }

        /**
         * @returns {boolean} whether this model is not persisted yet, ie. has not
         *                    been created yet. The default test is to check if the
         *                    model's identifier is missing.
         */

    }, {
        key: 'isNew',
        value: function isNew() {
            return _.isNil(this.identifier());
        }

        /**
         * @returns {boolean} the opposite of `isNew`, returns `true` if this model
         *                    is already persisted somewhere else.
         */

    }, {
        key: 'isExisting',
        value: function isExisting() {
            return !this.isNew();
        }

        /**
         * Called before a save request is made.
         *
         * @returns {boolean} `false` if the request should not be made.
         */

    }, {
        key: 'onSave',
        value: function onSave() {

            // Don't save if we're already busy saving this model.
            // This prevents things like accidental double-clicks.
            if (this.saving) {
                return false;
            }

            // Don't save if no data has changed, but consider it a success.
            if (!this.getOption('saveUnchanged') && !this.changed()) {
                return true;
            }

            // Mutate attribute before we save if required to do so.
            if (this.getOption('mutateBeforeSave')) {
                this.mutate();
            }

            // Validate all attributes before saving.
            if (!this.validate()) {
                throw new ValidationError(this.errors);
            }

            Vue.set(this, 'saving', true);
        }

        /**
         * Called before a delete request is made.
         *
         * @returns {boolean} `false` if the request should not be made.
         */

    }, {
        key: 'onDelete',
        value: function onDelete() {

            // Don't save if we're already busy deleting this model.
            if (this.deleting) {
                return false;
            }

            Vue.set(this, 'deleting', true);
        }
    }]);
    return Model;
}(Base);

/**
 * Models and Collections for Vue.js
 *
 * @version 0.2.3
 *
 * @author Rudi Theunissen <rudi.theunissen@figured.com>
 */

exports.Model = Model;
exports.Collection = Collection;
